//
//  MQSpeextDecoder.m
//  SpeexCodec
//
//  Created by cliftoncraig07 on 3/16/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "MySpeexDecoder.h"
#include <speex/speex_header.h>

#define MAX_FRAME_SIZE 2000
#if !defined(__LITTLE_ENDIAN__) && ( defined(WORDS_BIGENDIAN) || defined(__BIG_ENDIAN__) )
#define le_short(s) ((short) ((unsigned short) (s) << 8) | ((unsigned short) (s) >> 8))
#define be_short(s) ((short) (s))
#else
#define le_short(s) ((short) (s))
#define be_short(s) ((short) ((unsigned short) (s) << 8) | ((unsigned short) (s) >> 8))
#endif 

@implementation MySpeexDecoder
@synthesize delegate;
@synthesize numberOfDataReadPasses;

-(void) initializeOGGAndSpeex {
	/*Init Ogg data struct*/
	ogg_sync_init(&synchState);
	
	speex_bits_init(&bits);
	//narrow band == 0, wideband ==1, wide-wide-band? == 2
	theSpeexDecoder = speex_decoder_init(speex_lib_get_mode(1));
}

-(int) read:(int)numberOfBytesToRead bytesOfSpeexDataInBuffer:(char*)speexData {
	int numBytesRead = [dataStream read:(uint8_t*) speexData maxLength:numberOfBytesToRead];
	continueToDecode = [dataStream streamStatus]!=NSStreamStatusAtEnd;
	numberOfDataReadPasses++;
	return numBytesRead;
}

-(void) copyOggPacketToSpeexBitstream
{
	/*Copy Ogg packet to Speex bitstream*/
	speex_bits_read_from(&bits, (char*)oggPacket.packet, oggPacket.bytes);
	int numberOfFrames = [[audioAttributes valueForKey:@"framesPerPacket"] intValue];
	for (int i=0;i!=numberOfFrames;i++) {
		short int decodedData[MAX_FRAME_SIZE];
		speex_decode_int(theSpeexDecoder, &bits, decodedData);
		int blockSize = [[audioAttributes valueForKey:@"decodeBlockSize"] intValue];
		short int rawBlock[blockSize];
		for (int count=0; count < blockSize; count++) {
			rawBlock[count] = le_short( decodedData[count] );
		}
		if ([delegate respondsToSelector:@selector(audioWasDecoded:)]) {
			NSData *delegatesData = [NSData dataWithBytes:rawBlock length:blockSize];
			[delegate performSelector:@selector(audioWasDecoded:) withObject:delegatesData];
		}
	}
}

-(void) processOggPage {
	if (oggStreamState.serialno!=ogg_page_serialno(&oggPage)) {
		ogg_stream_reset_serialno(&oggStreamState, ogg_page_serialno(&oggPage));
	}
	ogg_stream_pagein(&oggStreamState, &oggPage);
	for(int oggStreamResult = ogg_stream_packetout(&oggStreamState, &oggPacket); !oggPacket.e_o_s && oggStreamResult==1; oggStreamResult = ogg_stream_packetout(&oggStreamState, &oggPacket)){
		[self copyOggPacketToSpeexBitstream];
	}
}

-(NSDictionary*) audioAttributesFromHeader:(SpeexHeader*)header
{
	if(audioAttributes) {[audioAttributes release];} //Release attributes if we have'em
	audioAttributes = [NSDictionary dictionaryWithObjectsAndKeys: [NSNumber numberWithInt:header->rate], @"sampleRate",
					[NSNumber numberWithInt:header->frame_size], @"frameSize",
					[NSNumber numberWithInt:header->nb_channels], @"numberOfChannels",
					[NSNumber numberWithInt:header->nb_channels * header->frame_size], @"decodeBlockSize",
					[NSNumber numberWithInt:header->frames_per_packet], @"framesPerPacket",
					nil];
	return audioAttributes;
}

-(void) decodeAudioAtPath:(NSString*)speexAudioPath
{
	numberOfDataReadPasses=0;
	dataStream = [NSInputStream inputStreamWithFileAtPath:speexAudioPath];
	[dataStream open];
	continueToDecode = YES;
	[self initializeOGGAndSpeex];
	
	ogg_sync_wrote( &synchState, [self read:200 bytesOfSpeexDataInBuffer: ogg_sync_buffer(&synchState, 200)] );
	ogg_sync_pageout(&synchState, &oggPage);
	ogg_stream_init(&oggStreamState, ogg_page_serialno(&oggPage));
	ogg_stream_pagein(&oggStreamState, &oggPage);
	ogg_stream_packetout(&oggStreamState, &oggPacket);
	if ([delegate respondsToSelector:@selector(willDecodeAudioWithAttributes:)]) {
		[delegate willDecodeAudioWithAttributes: [self audioAttributesFromHeader:speex_packet_to_header((char*)oggPacket.packet, oggPacket.bytes)]];
	}
	spx_int32_t enh_enabled = 1;
	speex_decoder_ctl(theSpeexDecoder, SPEEX_SET_ENH, &enh_enabled);
	spx_int32_t rate = [[audioAttributes valueForKey:@"sampleRate"] intValue];
	speex_decoder_ctl(theSpeexDecoder, SPEEX_SET_SAMPLING_RATE, &rate);

	while(continueToDecode) {
		while (ogg_sync_pageout(&synchState, &oggPage) == 1) {
			[self processOggPage];
		}
		int numBytesRead = [self read:200 bytesOfSpeexDataInBuffer:ogg_sync_buffer(&synchState, 200)];
		ogg_sync_wrote(&synchState, numBytesRead);
	}
	[dataStream close];
}

@end
